from shipment_approval_main import ShipmentApprovalApp
from PIL import Image, ImageTk
import json
import customtkinter as ctk
import pandas as pd
import os
import subprocess
import threading
import customtkinter as ctk

if __name__ == "__main__":
    app = ShipmentApprovalApp()
    app.mainloop()
